#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from services.Dispatchers.dispatcher_creator import DispatcherCreator
from tests.test_utils.config import load_test_config


def test_create_dispatchers(event_loop):
    config = load_test_config()
    dispatchers = DispatcherCreator.create_dispatchers(config, event_loop)
    assert len(dispatchers) == 3  # creates all dispatchers but none is ready (no config for associated services)
    for dispatcher in dispatchers:
        assert not dispatcher.get_is_setup_correctly()
